@@ -3,19 +3,72 @@ import argparse
 import asyncio
 import logging
 import re
-import typing
 from dataclasses import dataclass
+from typing import Awaitable, Callable, Dict, List, Optional, Union
 
 import paho.mqtt.client as mqtt
 import rhasspyhermes.cli as hermes_cli
 from rhasspyhermes.client import HermesClient
-from rhasspyhermes.dialogue import DialogueContinueSession, DialogueEndSession
+from rhasspyhermes.dialogue import (
+    DialogueContinueSession,
+    DialogueEndSession,
+    DialogueIntentNotRecognized,
+    DialogueNotification,
+    DialogueStartSession,
+)
 from rhasspyhermes.nlu import NluIntent, NluIntentNotRecognized
 from rhasspyhermes.wake import HotwordDetected
 
 _LOGGER = logging.getLogger("HermesApp")
 
 
+@dataclass
+class ContinueSession:
+    """Helper class to continue the current session.
+
+    Attributes:
+        text: The text the TTS should say to start this additional request of the session.
+        intent_filter: A list of intents names to restrict the NLU resolution on the
+            answer of this query.
+        custom_data: An update to the session's custom data. If not provided, the custom data
+            will stay the same.
+        send_intent_not_recognized: Indicates whether the dialogue manager should handle non recognized
+            intents by itself or send them for the client to handle.
+    """
+
+    custom_data: Optional[str] = None
+    text: Optional[str] = None
+    intent_filter: Optional[List[str]] = None
+    send_intent_not_recognized: bool = False
+
+
+@dataclass
+class EndSession:
+    """Helper class to end the current session.
+
+    Attributes:
+        text: The text the TTS should say to end the session.
+        custom_data: An update to the session's custom data. If not provided, the custom data
+            will stay the same.
+    """
+
+    text: Optional[str] = None
+    custom_data: Optional[str] = None
+
+
+@dataclass
+class TopicData:
+    """Helper class for topic subscription.
+
+    Attributes:
+        topic: The MQTT topic.
+        data: A dictionary holding extracted data for the given placeholder.
+    """
+
+    topic: str
+    data: Dict[str, str]
+
+
 class HermesApp(HermesClient):
     """A Rhasspy app using the Hermes protocol.
 
@@ -24,14 +77,14 @@ class HermesApp(HermesClient):
 
     Example:
 
-    .. literalinclude:: ../time_app.py
+    .. literalinclude:: ../examples/time_app.py
     """
 
     def __init__(
         self,
         name: str,
-        parser: typing.Optional[argparse.ArgumentParser] = None,
-        mqtt_client: typing.Optional[mqtt.Client] = None,
+        parser: Optional[argparse.ArgumentParser] = None,
+        mqtt_client: Optional[mqtt.Client] = None,
     ):
         """Initialize the Rhasspy Hermes app.
 
@@ -64,32 +117,34 @@ class HermesApp(HermesClient):
         # Initialize HermesClient
         super().__init__(name, mqtt_client, site_ids=self.args.site_id)
 
-        self._callbacks_hotword: typing.List[
-            typing.Callable[[HotwordDetected], None]
-        ] = []
+        self._callbacks_hotword: List[Callable[[HotwordDetected], Awaitable[None]]] = []
 
-        self._callbacks_intent: typing.Dict[
-            str, typing.List[typing.Callable[[NluIntent], None]],
+        self._callbacks_intent: Dict[
+            str, List[Callable[[NluIntent], Awaitable[None]]],
         ] = {}
 
-        self._callbacks_intent_not_recognized: typing.List[
-            typing.Callable[[NluIntentNotRecognized], None]
+        self._callbacks_intent_not_recognized: List[
+            Callable[[NluIntentNotRecognized], Awaitable[None]]
         ] = []
 
-        self._callbacks_topic: typing.Dict[
-            str, typing.List[typing.Callable[[TopicData, bytes], None]]
+        self._callbacks_dialogue_intent_not_recognized: List[
+            Callable[[DialogueIntentNotRecognized], Awaitable[None]]
+        ] = []
+
+        self._callbacks_topic: Dict[
+            str, List[Callable[[TopicData, bytes], Awaitable[None]]]
         ] = {}
 
-        self._callbacks_topic_regex: typing.List[
-            typing.Callable[[TopicData, bytes], None]
+        self._callbacks_topic_regex: List[
+            Callable[[TopicData, bytes], Awaitable[None]]
         ] = []
 
-        self._additional_topic: typing.List[str] = []
+        self._additional_topic: List[str] = []
 
     def _subscribe_callbacks(self) -> None:
         # Remove duplicate intent names
-        intent_names: typing.List[str] = list(set(self._callbacks_intent.keys()))
-        topics: typing.List[str] = [
+        intent_names: List[str] = list(set(self._callbacks_intent.keys()))
+        topics: List[str] = [
             NluIntent.topic(intent_name=intent_name) for intent_name in intent_names
         ]
 
@@ -99,7 +154,10 @@ class HermesApp(HermesClient):
         if self._callbacks_intent_not_recognized:
             topics.append(NluIntentNotRecognized.topic())
 
-        topic_names: typing.List[str] = list(set(self._callbacks_topic.keys()))
+        if self._callbacks_dialogue_intent_not_recognized:
+            topics.append(DialogueIntentNotRecognized.topic())
+
+        topic_names: List[str] = list(set(self._callbacks_topic.keys()))
         topics.extend(topic_names)
         topics.extend(self._additional_topic)
 
@@ -121,7 +179,7 @@ class HermesApp(HermesClient):
                 try:
                     hotword_detected = HotwordDetected.from_json(payload)
                     for function_h in self._callbacks_hotword:
-                        function_h(hotword_detected)
+                        await function_h(hotword_detected)
                 except KeyError as key:
                     _LOGGER.error(
                         "Missing key %s in JSON payload for %s: %s", key, topic, payload
@@ -133,7 +191,7 @@ class HermesApp(HermesClient):
                     intent_name = nlu_intent.intent.intent_name
                     if intent_name in self._callbacks_intent:
                         for function_i in self._callbacks_intent[intent_name]:
-                            function_i(nlu_intent)
+                            await function_i(nlu_intent)
                 except KeyError as key:
                     _LOGGER.error(
                         "Missing key %s in JSON payload for %s: %s", key, topic, payload
@@ -145,7 +203,19 @@ class HermesApp(HermesClient):
                         payload
                     )
                     for function_inr in self._callbacks_intent_not_recognized:
-                        function_inr(nlu_intent_not_recognized)
+                        await function_inr(nlu_intent_not_recognized)
+                except KeyError as key:
+                    _LOGGER.error(
+                        "Missing key %s in JSON payload for %s: %s", key, topic, payload
+                    )
+            elif DialogueIntentNotRecognized.is_topic(topic):
+                # hermes/dialogueManager/intentNotRecognized
+                try:
+                    dialogue_intent_not_recognized = DialogueIntentNotRecognized.from_json(
+                        payload
+                    )
+                    for function_dinr in self._callbacks_dialogue_intent_not_recognized:
+                        await function_dinr(dialogue_intent_not_recognized)
                 except KeyError as key:
                     _LOGGER.error(
                         "Missing key %s in JSON payload for %s: %s", key, topic, payload
@@ -154,7 +224,7 @@ class HermesApp(HermesClient):
                 unexpected_topic = True
                 if topic in self._callbacks_topic:
                     for function_1 in self._callbacks_topic[topic]:
-                        function_1(TopicData(topic, {}), payload)
+                        await function_1(TopicData(topic, {}), payload)
                         unexpected_topic = False
                 else:
                     for function_2 in self._callbacks_topic_regex:
@@ -178,8 +248,8 @@ class HermesApp(HermesClient):
             _LOGGER.exception("on_raw_message")
 
     def on_hotword(
-        self, function: typing.Callable[[HotwordDetected], None]
-    ) -> typing.Callable[[HotwordDetected], None]:
+        self, function: Callable[[HotwordDetected], Awaitable[None]]
+    ) -> Callable[[HotwordDetected], Awaitable[None]]:
         """Apply this decorator to a function that you want to act on a detected hotword.
 
         The decorated function has a :class:`rhasspyhermes.wake.HotwordDetected` object as an argument
@@ -190,7 +260,7 @@ class HermesApp(HermesClient):
         .. code-block:: python
 
             @app.on_hotword
-            def wake(hotword: HotwordDetected):
+            async def wake(hotword: HotwordDetected):
                 print(f"Hotword {hotword.model_id} detected on site {hotword.site_id}")
 
         If a hotword has been detected, the ``wake`` function is called with the ``hotword`` argument.
@@ -203,9 +273,13 @@ class HermesApp(HermesClient):
 
     def on_intent(
         self, *intent_names: str
-    ) -> typing.Callable[
-        [typing.Callable[[NluIntent], typing.Union["ContinueSession", "EndSession"]]],
-        typing.Callable[[NluIntent], None],
+    ) -> Callable[
+        [
+            Callable[
+                [NluIntent], Union[Awaitable[ContinueSession], Awaitable[EndSession]]
+            ]
+        ],
+        Callable[[NluIntent], Awaitable[None]],
     ]:
         """Apply this decorator to a function that you want to act on a received intent.
 
@@ -224,7 +298,7 @@ class HermesApp(HermesClient):
         .. code-block:: python
 
             @app.on_intent("GetTime")
-            def get_time(intent: NluIntent):
+            async def get_time(intent: NluIntent):
                 return EndSession("It's too late.")
 
         If the intent with name GetTime has been detected, the ``get_time`` function is called
@@ -232,12 +306,12 @@ class HermesApp(HermesClient):
         """
 
         def wrapper(
-            function: typing.Callable[
-                [NluIntent], typing.Union[ContinueSession, EndSession]
+            function: Callable[
+                [NluIntent], Union[Awaitable[ContinueSession], Awaitable[EndSession]]
             ]
-        ) -> typing.Callable[[NluIntent], None]:
-            def wrapped(intent: NluIntent) -> None:
-                message = function(intent)
+        ) -> Callable[[NluIntent], Awaitable[None]]:
+            async def wrapped(intent: NluIntent) -> None:
+                message = await function(intent)
                 if isinstance(message, EndSession):
                     if intent.session_id is not None:
                         self.publish(
@@ -281,11 +355,11 @@ class HermesApp(HermesClient):
 
     def on_intent_not_recognized(
         self,
-        function: typing.Callable[
+        function: Callable[
             [NluIntentNotRecognized],
-            typing.Union["ContinueSession", "EndSession", None],
+            Union[Awaitable[ContinueSession], Awaitable[EndSession], Awaitable[None]],
         ],
-    ) -> typing.Callable[[NluIntentNotRecognized], None]:
+    ) -> Callable[[NluIntentNotRecognized], Awaitable[None]]:
         """Apply this decorator to a function that you want to act when the NLU system
         hasn't recognized an intent.
 
@@ -302,15 +376,15 @@ class HermesApp(HermesClient):
         .. code-block:: python
 
             @app.on_intent_not_recognized
-            def not_understood(intent_not_recognized: NluIntentNotRecognized):
+            async def not_understood(intent_not_recognized: NluIntentNotRecognized):
                 print(f"Didn't understand \"{intent_not_recognized.input}\" on site {intent_not_recognized.site_id}")
 
         If an intent hasn't been recognized, the ``not_understood`` function is called
         with the ``intent_not_recognized`` argument. This object holds information about the not recognized intent.
         """
 
-        def wrapped(inr: NluIntentNotRecognized) -> None:
-            message = function(inr)
+        async def wrapped(inr: NluIntentNotRecognized) -> None:
+            message = await function(inr)
             if isinstance(message, EndSession):
                 if inr.session_id is not None:
                     self.publish(
@@ -323,7 +397,7 @@ class HermesApp(HermesClient):
                     )
                 else:
                     _LOGGER.error(
-                        "Cannot end session of intent not recognized message without session ID."
+                        "Cannot end session of NLU intent not recognized message without session ID."
                     )
             elif isinstance(message, ContinueSession):
                 if inr.session_id is not None:
@@ -339,13 +413,81 @@ class HermesApp(HermesClient):
                     )
                 else:
                     _LOGGER.error(
-                        "Cannot continue session of intent not recognized message without session ID."
+                        "Cannot continue session of NLU intent not recognized message without session ID."
                     )
 
         self._callbacks_intent_not_recognized.append(wrapped)
 
         return wrapped
 
+    def on_dialogue_intent_not_recognized(
+        self,
+        function: Callable[
+            [DialogueIntentNotRecognized],
+            Union[Awaitable[ContinueSession], Awaitable[EndSession], Awaitable[None]],
+        ],
+    ) -> Callable[[DialogueIntentNotRecognized], Awaitable[None]]:
+        """Apply this decorator to a function that you want to act when the dialogue manager
+        failed to recognize an intent and you requested to notify you of this event with the
+        `sendIntentNotRecognized` flag.
+
+        The decorated function has a :class:`rhasspyhermes.dialogue.DialogueIntentNotRecognized` object as an argument
+        and can return a :class:`ContinueSession` or :class:`EndSession` object or have no return value.
+
+        If the function returns a :class:`ContinueSession` object, the current session is continued after
+        saying the supplied text. If the function returns a a :class:`EndSession` object, the current session
+        is ended after saying the supplied text, or immediately when no text is supplied. If the function doesn't
+        have a return value, nothing is changed to the session.
+
+        Example:
+
+        .. code-block:: python
+
+            @app.on_dialogue_intent_not_recognized
+            async def not_understood(intent_not_recognized: DialogueIntentNotRecognized):
+                print(f"Didn't understand \"{intent_not_recognized.input}\" on site {intent_not_recognized.site_id}")
+
+        If an intent hasn't been recognized, the ``not_understood`` function is called
+        with the ``intent_not_recognized`` argument. This object holds information about the not recognized intent.
+        """
+
+        async def wrapped(inr: DialogueIntentNotRecognized) -> None:
+            message = await function(inr)
+            if isinstance(message, EndSession):
+                if inr.session_id is not None:
+                    self.publish(
+                        DialogueEndSession(
+                            session_id=inr.session_id,
+                            site_id=inr.site_id,
+                            text=message.text,
+                            custom_data=message.custom_data,
+                        )
+                    )
+                else:
+                    _LOGGER.error(
+                        "Cannot end session of dialogue intent not recognized message without session ID."
+                    )
+            elif isinstance(message, ContinueSession):
+                if inr.session_id is not None:
+                    self.publish(
+                        DialogueContinueSession(
+                            session_id=inr.session_id,
+                            site_id=inr.site_id,
+                            text=message.text,
+                            intent_filter=message.intent_filter,
+                            custom_data=message.custom_data,
+                            send_intent_not_recognized=message.send_intent_not_recognized,
+                        )
+                    )
+                else:
+                    _LOGGER.error(
+                        "Cannot continue session of dialogue intent not recognized message without session ID."
+                    )
+
+        self._callbacks_dialogue_intent_not_recognized.append(wrapped)
+
+        return wrapped
+
     def on_topic(self, *topic_names: str):
         """Apply this decorator to a function that you want to act on a received raw MQTT message.
 
@@ -360,7 +502,7 @@ class HermesApp(HermesClient):
         .. code-block:: python
 
             @app.on_topic("hermes/+/{site_id}/playBytes/#")
-            def test_topic1(data: TopicData, payload: bytes):
+            async def test_topic1(data: TopicData, payload: bytes):
                 _LOGGER.debug("topic: %s, site_id: %s", data.topic, data.data.get("site_id"))
 
         .. note:: The topic names can contain MQTT wildcards (`+` and `#`) or templates (`{foobar}`).
@@ -369,8 +511,8 @@ class HermesApp(HermesClient):
         """
 
         def wrapper(function):
-            def wrapped(data: TopicData, payload: bytes):
-                function(data, payload)
+            async def wrapped(data: TopicData, payload: bytes):
+                await function(data, payload)
 
             replaced_topic_names = []
 
@@ -463,49 +605,14 @@ class HermesApp(HermesClient):
         finally:
             self.mqtt_client.loop_stop()
 
+    def notify(self, text: str, site_id: str = "default"):
+        """Send a dialogue notification.
 
-@dataclass
-class ContinueSession:
-    """Helper class to continue the current session.
+        Use this to inform the user of something without expecting a response.
 
-    Attributes:
-        text: The text the TTS should say to start this additional request of the session.
-        intent_filter: A list of intents names to restrict the NLU resolution on the
-            answer of this query.
-        custom_data: An update to the session's custom data. If not provided, the custom data
-            will stay the same.
-        send_intent_not_recognized: Indicates whether the dialogue manager should handle non recognized
-            intents by itself or send them for the client to handle.
-    """
-
-    custom_data: typing.Optional[str] = None
-    text: typing.Optional[str] = None
-    intent_filter: typing.Optional[typing.List[str]] = None
-    send_intent_not_recognized: bool = False
-
-
-@dataclass
-class EndSession:
-    """Helper class to end the current session.
-
-    Attributes:
-        text: The text the TTS should say to end the session.
-        custom_data: An update to the session's custom data. If not provided, the custom data
-            will stay the same.
-    """
-
-    text: typing.Optional[str] = None
-    custom_data: typing.Optional[str] = None
-
-
-@dataclass
-class TopicData:
-    """Helper class for topic subscription.
-
-    Attributes:
-        topic: The MQTT topic.
-        data: A dictionary holding extracted data for the given placeholder.
-    """
-
-    topic: str
-    data: typing.Dict[str, str]
+        Arguments:
+            text: The text to say.
+            site_id: The ID of the site where the text should be said.
+        """
+        notification = DialogueNotification(text)
+        self.publish(DialogueStartSession(init=notification, site_id=site_id))
